//******* randbeta.c		by Luigi Marino, 2010 - Updated and recompiled in 2017 for Max 7 at 32 and 64 bit

#include "ext.h"							// standard Max include, always required
#include "ext_obex.h"						// required for new style Max object
#define MY_RAND_MAX ((1U << 31) - 1)		

////////////////////////// object struct
typedef struct _randbeta
	{
		t_object					r_ob;			
		double						r_min;
		double						r_max;
		double						r_bias;
		double						r_alpha;
		unsigned int				r_rand;		//this variable keeps track of the random series on each single instance 
		void						*r_out;
	} t_randbeta;

///////////////////////// function prototypes
//// standard set
void *randbeta_new(t_symbol *s, long argc, t_atom *argv);
void randbeta_assist(t_randbeta *x, void *b, long m, long a, char *s);
//// additional methods
void randbeta_bang(t_randbeta *x); 
void randbeta_ft1(t_randbeta *x, double f);
void randbeta_ft2(t_randbeta *x, double f); 
void randbeta_ft3(t_randbeta *x, double f);
void randbeta_ft4(t_randbeta *x, double f); 
void randbeta_seed(t_randbeta *x, long n);
//// other functions
double deviation_float(double cval, double dev, double alpha);
long deviation_int(long cval, long dev, double alpha);
double reflect_float(double f, double min, double max);
long reflect_int(long n, long min, long max);
short even(long n);
long myround(double f);
short coin(t_randbeta *x, double f);
int myrand(t_randbeta *x);
void mysrand(t_randbeta *x, unsigned int i);
double pcauchy(t_randbeta *x, double alpha);
double ncauchy(t_randbeta *x, double alpha);
double sortOfBeta(t_randbeta *x, double min, double max, double bias, double spread);
//////////////////////// global pointer to the object class - so max can reference the object
t_class *randbeta_class;

void ext_main(void *r)
{
	t_class *c;
	
	c = class_new("randbeta", (method)randbeta_new, (method)NULL, (long)sizeof(t_randbeta), 0L, A_GIMME, 0);
	
    class_addmethod(c, (method)randbeta_assist, "assist", A_CANT, 0);  
	class_addmethod(c, (method)randbeta_bang, "bang", A_FLOAT, 0);
	class_addmethod(c, (method)randbeta_ft1, "ft1", A_FLOAT, 0);
	class_addmethod(c, (method)randbeta_ft2, "ft2", A_FLOAT, 0);
	class_addmethod(c, (method)randbeta_ft3, "ft3", A_FLOAT, 0);
	class_addmethod(c, (method)randbeta_ft4, "ft4", A_FLOAT, 0);	
	class_addmethod(c, (method)randbeta_seed, "seed", A_LONG, 0);
	
	class_register(CLASS_BOX, c);
	randbeta_class = c;
	
}

void randbeta_assist(t_randbeta *x, void *b, long m, long a, char *s)
{
	if (m == ASSIST_INLET) {
		switch (a) {
			case 0:
				sprintf(s, "bang generates the value");
				break;
			case 1:
				sprintf(s, "min - float");
				break;
			case 2:
				sprintf(s, "max - float");
				break;
			case 3:
				sprintf(s, "bias - float[0,1]");
				break;
			case 4:
				sprintf(s, "distribution - float[0,1]");
				break;				
		}
	} else {	
		sprintf(s, "output");
	}
}



void randbeta_bang(t_randbeta *x)
{
	double out;
	out = sortOfBeta(x, x->r_min, x->r_max, x->r_bias, x->r_alpha); 
	outlet_float(x->r_out, out);
}



void randbeta_ft1(t_randbeta *x, double f)
{
	x->r_min = f;
}

void randbeta_ft2(t_randbeta *x, double f)
{
	x->r_max = f;
}

void randbeta_ft3(t_randbeta *x, double f)
{
	if (f < 0.)
		f = 0.;
	else if (f > 1.)
		f = 1.;
	else
		;
			 
	x->r_bias = f;
}

void randbeta_ft4(t_randbeta *x, double f)
{
	f = fabs(f);
	x->r_alpha = f;
}

void randbeta_seed(t_randbeta *x, long n)
{
	mysrand(x, (unsigned int)n);
}

static int makeseed(void)
{
	static unsigned int random_nextseed = 1489853723;
	random_nextseed = random_nextseed * 435898247 + 938284287;
	return (random_nextseed & 0x7fffffff);
}		

void mysrand(t_randbeta *x, unsigned int n)  
{	
	if (n == 0) {
		x->r_rand = makeseed();  
	} else
		x->r_rand = n;
	/*
	In order to initialize randomly the series, it is usually used the system clock as seed. 
	Doing so, if you have multiple instances of a random generator on the same patch they'll be initialized at the same time getting the same seed.
	makeseed() is declared as static so that when it is called by multiple instances at the same time you get different seeds.
	*/
}

int myrand(t_randbeta *x) //Linear Congruential Generator (LCG) 
{
	return (x->r_rand = (x->r_rand * 1103515245 + 12345) & MY_RAND_MAX);
	/*
	Implementation of rand() rewritten using a variable in the object structure to allow a separate random series for each object instance. 
	The quality of a LCG depends a lot on the multiplier and increment (here 1103515245 and 12345).
	For other methods to generate random numbers check out Mersenne Twister or R250/521.  
	*/						
}

double cauchy(t_randbeta *x, double cval, double alpha)
{

	double uni, cau;
	const double PI = 3.14159265358979; 
				
	if (alpha == 0.0) { 
		cau = cval;	
	} else {
		do
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 0.5);
		cau = alpha * (tan(PI * uni));
	}

	return cau;
}

double pcauchy(t_randbeta *x, double alpha)
{
	
	double uni, pcau;
	const double PI = 3.14159265358979; 
	
	if (alpha == 0.0) { 
		pcau = 0.;	
	} else {
		do
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 1.);
		pcau = alpha * (tan((PI/2.) * uni));
	}
	
	return pcau;
}

double ncauchy(t_randbeta *x, double alpha)
{
	
	double uni, ncau;
	const double PI = 3.14159265358979; 
	
	if (alpha == 0.0) { 
		ncau = 0.;	
	} else {
		do
			uni = (double)myrand(x) / MY_RAND_MAX;     
		while (uni == 1.);
		ncau = alpha * tan(((PI * uni) + PI) / 2.);
	}
	
	return ncau;
}


double sortOfBeta(t_randbeta *x, double min, double max, double bias, double spread)
{
	double alpha, sbeta;
	
	alpha = spread; 
	
	if (spread < 1.) {
		if (coin(x, bias))
			sbeta = reflect_float(ncauchy(x, alpha), -1., 0.) + 1;
		else
			sbeta = reflect_float(pcauchy(x, alpha), 0., 1.);
	} else {
		sbeta = (double)myrand(x) / MY_RAND_MAX;
	}
	
	return ((max - min) * sbeta) + min; 

}
	


double reflect_float(double f, double min, double max)
{
	double rng, frel;
	unsigned long c;
	
	rng = fabs(max - min);
	
	if (rng == 0) 
		return min;
	
	frel = f - min;
	c = (long)(frel / rng);
	
	if (!even(c)) 
		frel = rng - fabs(fmod(frel, rng));
	else
		frel = fabs(fmod(frel, rng));
	
	f = frel + min;
	
	return f;
}	

short even(long n)
{
	if ((n % 2) == 0) 
		return 1;
	else 
		return 0;
}

long myround(double f)
{
	long n;
	if (f >= 0) {
		if (fmod(f, 1.0) >= 0.5)
			n = (long)f + 1;
		else
			n = (long)f;
	} else {
		if (fmod(f, 1.0) <= -0.5)
			n = (long)f - 1;
		else
			n = (long)f;
	}
	return n;
}	

short coin(t_randbeta *x, double f)
{ 				
	double i;
	
	i = (double)myrand(x) / MY_RAND_MAX;	
	
	if (i <= f) 
		return 1;
	else
		return 0;
}



void *randbeta_new(t_symbol *s, long argc, t_atom *argv)
{
	t_randbeta *x = NULL;
    
	x = (t_randbeta *)object_alloc(randbeta_class); 
	
	switch (argc) {
		case 0:
			x->r_min = 0.;
			x->r_max = 1.;
			x->r_bias = 0.5;
			x->r_alpha = 0.05;
			break;
		case 1:
			x->r_min = 0.;
			x->r_max = 1.;
			x->r_bias = 0.5;
			x->r_alpha = atom_getfloat(argv);
			break;
		case 2:
			x->r_min = 0.;
			x->r_max = 1.;
			x->r_bias = atom_getfloat(argv);
			x->r_alpha = atom_getfloat(argv+1);
			break;			
		case 4:
			x->r_min = atom_getfloat(argv);
			x->r_max = atom_getfloat(argv+1);
			x->r_bias = atom_getfloat(argv+2);
			x->r_alpha = atom_getfloat(argv+3);
			break;			
		default:
			object_warn((t_object *)x, "wrong arguments, default initialization");
			x->r_min = 0.;
			x->r_max = 1.;
			x->r_bias = 0.5;
			x->r_alpha = 0.05;
			break;
	}
	
	mysrand(x, 0);
	
	floatin(x,4);	
	floatin(x,3);								  
	floatin(x,2);	
	floatin(x,1);
	
	x->r_out = outlet_new(x, NULL);
		
	
	return(x);
}

